@@ -64,8 +64,8 @@ int ares_parse_mx_reply(const unsigned char *abuf, int alen_int,
 
     if (rr == NULL) {
       /* Shouldn't be possible */
-      status = ARES_EBADRESP;
-      goto done;
+      status = ARES_EBADRESP; /* LCOV_EXCL_LINE: DefensiveCoding */
+      goto done; /* LCOV_EXCL_LINE: DefensiveCoding */
     }
 
     if (ares_dns_rr_get_class(rr) != ARES_CLASS_IN ||
@@ -76,8 +76,8 @@ int ares_parse_mx_reply(const unsigned char *abuf, int alen_int,
     /* Allocate storage for this MX answer appending it to the list */
     mx_curr = ares_malloc_data(ARES_DATATYPE_MX_REPLY);
     if (mx_curr == NULL) {
-      status = ARES_ENOMEM;
-      goto done;
+      status = ARES_ENOMEM; /* LCOV_EXCL_LINE: OutOfMemory */
+      goto done; /* LCOV_EXCL_LINE: OutOfMemory */
     }
 
     /* Link in the record */
@@ -92,8 +92,8 @@ int ares_parse_mx_reply(const unsigned char *abuf, int alen_int,
     mx_curr->host = ares_strdup(ares_dns_rr_get_str(rr, ARES_RR_MX_EXCHANGE));
 
     if (mx_curr->host == NULL) {
-      status = ARES_ENOMEM;
-      goto done;
+      status = ARES_ENOMEM; /* LCOV_EXCL_LINE: OutOfMemory */
+      goto done; /* LCOV_EXCL_LINE: OutOfMemory */
     }
   }
 
