@@ -5,8 +5,9 @@
 #ifndef V8_OBJECTS_JS_PROXY_INL_H_
 #define V8_OBJECTS_JS_PROXY_INL_H_
 
+#include "src/objects/instance-type-inl.h"
+#include "src/objects/js-objects-inl.h"
 #include "src/objects/js-proxy.h"
-
 #include "src/objects/objects-inl.h"  // Needed for write barriers
 
 // Has to be the last include (doesn't have include guards):
@@ -15,6 +16,8 @@
 namespace v8 {
 namespace internal {
 
+#include "torque-generated/src/objects/js-proxy-tq-inl.inc"
+
 TQ_OBJECT_CONSTRUCTORS_IMPL(JSProxy)
 
 bool JSProxy::IsRevoked() const { return !handler().IsJSReceiver(); }
