@@ -540,11 +540,14 @@ primordials.SafePromiseAllReturnVoid = (promises, mapFn) =>
   new Promise((resolve, reject) => {
     let pendingPromises = promises.length;
     if (pendingPromises === 0) resolve();
+    const onFulfilled = () => {
+      if (--pendingPromises === 0) {
+        resolve();
+      }
+    };
     for (let i = 0; i < promises.length; i++) {
       const promise = mapFn != null ? mapFn(promises[i], i) : promises[i];
-      PromisePrototypeThen(PromiseResolve(promise), () => {
-        if (--pendingPromises === 0) resolve();
-      }, reject);
+      PromisePrototypeThen(PromiseResolve(promise), onFulfilled, reject);
     }
   });
 
@@ -569,9 +572,17 @@ primordials.SafePromiseAllSettled = (promises, mapFn) =>
  * @param {(v: T|PromiseLike<T>, k: number) => U|PromiseLike<U>} [mapFn]
  * @returns {Promise<void>}
  */
-primordials.SafePromiseAllSettledReturnVoid = async (promises, mapFn) => {
-  await primordials.SafePromiseAllSettled(promises, mapFn);
-};
+primordials.SafePromiseAllSettledReturnVoid = (promises, mapFn) => new Promise((resolve) => {
+  let pendingPromises = promises.length;
+  if (pendingPromises === 0) resolve();
+  const onSettle = () => {
+    if (--pendingPromises === 0) resolve();
+  };
+  for (let i = 0; i < promises.length; i++) {
+    const promise = mapFn != null ? mapFn(promises[i], i) : promises[i];
+    PromisePrototypeThen(PromiseResolve(promise), onSettle, onSettle);
+  }
+});
 
 /**
  * @template T,U
