@@ -6,6 +6,7 @@ const {
   JSONParse,
   MathFloor,
   NumberParseInt,
+  ObjectAssign,
   RegExpPrototypeExec,
   RegExpPrototypeSymbolSplit,
   SafeMap,
@@ -13,6 +14,8 @@ const {
   StringPrototypeIncludes,
   StringPrototypeLocaleCompare,
   StringPrototypeStartsWith,
+  MathMax,
+  MathMin,
 } = primordials;
 const {
   copyFileSync,
@@ -24,7 +27,8 @@ const {
 const { setupCoverageHooks } = require('internal/util');
 const { tmpdir } = require('os');
 const { join, resolve } = require('path');
-const { fileURLToPath } = require('url');
+const { fileURLToPath } = require('internal/url');
+const { kMappings, SourceMap } = require('internal/source_map/source_map');
 const kCoverageFileRegex = /^coverage-(\d+)-(\d{13})-(\d+)\.json$/;
 const kIgnoreRegex = /\/\* node:coverage ignore next (?<count>\d+ )?\*\//;
 const kLineEndingRegex = /\r?\n$/u;
@@ -32,31 +36,16 @@ const kLineSplitRegex = /(?<=\r?\n)/u;
 const kStatusRegex = /\/\* node:coverage (?<status>enable|disable) \*\//;
 
 class CoverageLine {
-  #covered;
-
-  constructor(line, src, startOffset) {
-    const newlineLength =
+  constructor(line, startOffset, src, length = src?.length) {
+    const newlineLength = src == null ? 0 :
       RegExpPrototypeExec(kLineEndingRegex, src)?.[0].length ?? 0;
 
     this.line = line;
     this.src = src;
     this.startOffset = startOffset;
-    this.endOffset = startOffset + src.length - newlineLength;
+    this.endOffset = startOffset + length - newlineLength;
     this.ignore = false;
-    this.#covered = true;
-  }
-
-  get covered() {
-    return this.#covered;
-  }
-
-  set covered(isCovered) {
-    // V8 can generate multiple ranges that span the same line.
-    if (!this.#covered) {
-      return;
-    }
-
-    this.#covered = isCovered;
+    this.count = this.startOffset === this.endOffset ? 1 : 0;
   }
 }
 
@@ -67,9 +56,74 @@ class TestCoverage {
     this.workingDirectory = workingDirectory;
   }
 
+  #sourceLines = new SafeMap();
+
+  getLines(fileUrl, source) {
+    // Split the file source into lines. Make sure the lines maintain their
+    // original line endings because those characters are necessary for
+    // determining offsets in the file.
+    if (this.#sourceLines.has(fileUrl)) {
+      return this.#sourceLines.get(fileUrl);
+    }
+
+    try {
+      source ??= readFileSync(fileURLToPath(fileUrl), 'utf8');
+    } catch {
+      // The file can no longer be read. It may have been deleted among
+      // other possibilities. Leave it out of the coverage report.
+      this.#sourceLines.set(fileUrl, null);
+      return;
+    }
+
+    const linesWithBreaks =
+      RegExpPrototypeSymbolSplit(kLineSplitRegex, source);
+    let ignoreCount = 0;
+    let enabled = true;
+    let offset = 0;
+
+    const lines = ArrayPrototypeMap(linesWithBreaks, (line, i) => {
+      const startOffset = offset;
+      const coverageLine = new CoverageLine(i + 1, startOffset, line);
+
+      offset += line.length;
+
+      // Determine if this line is being ignored.
+      if (ignoreCount > 0) {
+        ignoreCount--;
+        coverageLine.ignore = true;
+      } else if (!enabled) {
+        coverageLine.ignore = true;
+      }
+
+      if (!coverageLine.ignore) {
+        // If this line is not already being ignored, check for ignore
+        // comments.
+        const match = RegExpPrototypeExec(kIgnoreRegex, line);
+
+        if (match !== null) {
+          ignoreCount = NumberParseInt(match.groups?.count ?? 1, 10);
+        }
+      }
+
+      // Check for comments to enable/disable coverage no matter what. These
+      // take precedence over ignore comments.
+      const match = RegExpPrototypeExec(kStatusRegex, line);
+      const status = match?.groups?.status;
+
+      if (status) {
+        ignoreCount = 0;
+        enabled = status === 'enable';
+      }
+
+      return coverageLine;
+    });
+    this.#sourceLines.set(fileUrl, lines);
+    return lines;
+  }
+
   summary() {
     internalBinding('profiler').takeCoverage();
-    const coverage = getCoverageFromDirectory(this.coverageDirectory);
+    const coverage = this.getCoverageFromDirectory();
     const coverageSummary = {
       __proto__: null,
       workingDirectory: this.workingDirectory,
@@ -95,76 +149,38 @@ class TestCoverage {
     for (let i = 0; i < coverage.length; ++i) {
       const { functions, url } = coverage[i];
 
-      // Split the file source into lines. Make sure the lines maintain their
-      // original line endings because those characters are necessary for
-      // determining offsets in the file.
-      const filePath = fileURLToPath(url);
-      let source;
-
-      try {
-        source = readFileSync(filePath, 'utf8');
-      } catch {
-        // The file can no longer be read. It may have been deleted among
-        // other possibilities. Leave it out of the coverage report.
-        continue;
-      }
-
-      const linesWithBreaks =
-        RegExpPrototypeSymbolSplit(kLineSplitRegex, source);
-      let ignoreCount = 0;
-      let enabled = true;
-      let offset = 0;
       let totalBranches = 0;
       let totalFunctions = 0;
       let branchesCovered = 0;
       let functionsCovered = 0;
+      const functionReports = [];
+      const branchReports = [];
 
-      const lines = ArrayPrototypeMap(linesWithBreaks, (line, i) => {
-        const startOffset = offset;
-        const coverageLine = new CoverageLine(i + 1, line, startOffset);
+      const lines = this.getLines(url);
+      if (!lines) {
+        continue;
+      }
 
-        offset += line.length;
-
-        // Determine if this line is being ignored.
-        if (ignoreCount > 0) {
-          ignoreCount--;
-          coverageLine.ignore = true;
-        } else if (!enabled) {
-          coverageLine.ignore = true;
-        }
-
-        if (!coverageLine.ignore) {
-          // If this line is not already being ignored, check for ignore
-          // comments.
-          const match = RegExpPrototypeExec(kIgnoreRegex, line);
-
-          if (match !== null) {
-            ignoreCount = NumberParseInt(match.groups?.count ?? 1, 10);
-          }
-        }
-
-        // Check for comments to enable/disable coverage no matter what. These
-        // take precedence over ignore comments.
-        const match = RegExpPrototypeExec(kStatusRegex, line);
-        const status = match?.groups?.status;
-
-        if (status) {
-          ignoreCount = 0;
-          enabled = status === 'enable';
-        }
-
-        return coverageLine;
-      });
 
       for (let j = 0; j < functions.length; ++j) {
         const { isBlockCoverage, ranges } = functions[j];
 
+        let maxCountPerFunction = 0;
         for (let k = 0; k < ranges.length; ++k) {
           const range = ranges[k];
+          maxCountPerFunction = MathMax(maxCountPerFunction, range.count);
 
-          mapRangeToLines(range, lines);
+          // Add some useful data to the range. The test runner has read these ranges
+          // from a file, so we own the data structures and can do what we want.
+          ObjectAssign(range, mapRangeToLines(range, lines));
 
           if (isBlockCoverage) {
+            ArrayPrototypePush(branchReports, {
+              __proto__: null,
+              line: range.lines[0]?.line,
+              count: range.count,
+            });
+
             if (range.count !== 0 ||
                 range.ignoredLines === range.lines.length) {
               branchesCovered++;
@@ -177,6 +193,13 @@ class TestCoverage {
         if (j > 0 && ranges.length > 0) {
           const range = ranges[0];
 
+          ArrayPrototypePush(functionReports, {
+            __proto__: null,
+            name: functions[j].functionName,
+            count: maxCountPerFunction,
+            line: range.lines[0]?.line,
+          });
+
           if (range.count !== 0 || range.ignoredLines === range.lines.length) {
             functionsCovered++;
           }
@@ -186,21 +209,25 @@ class TestCoverage {
       }
 
       let coveredCnt = 0;
-      const uncoveredLineNums = [];
+      const lineReports = [];
 
       for (let j = 0; j < lines.length; ++j) {
         const line = lines[j];
-
-        if (line.covered || line.ignore) {
+        if (!line.ignore) {
+          ArrayPrototypePush(lineReports, {
+            __proto__: null,
+            line: line.line,
+            count: line.count,
+          });
+        }
+        if (line.count > 0 || line.ignore) {
           coveredCnt++;
-        } else {
-          ArrayPrototypePush(uncoveredLineNums, line.line);
         }
       }
 
       ArrayPrototypePush(coverageSummary.files, {
         __proto__: null,
-        path: filePath,
+        path: fileURLToPath(url),
         totalLineCount: lines.length,
         totalBranchCount: totalBranches,
         totalFunctionCount: totalFunctions,
@@ -210,7 +237,9 @@ class TestCoverage {
         coveredLinePercent: toPercentage(coveredCnt, lines.length),
         coveredBranchPercent: toPercentage(branchesCovered, totalBranches),
         coveredFunctionPercent: toPercentage(functionsCovered, totalFunctions),
-        uncoveredLineNumbers: uncoveredLineNums,
+        functions: functionReports,
+        branches: branchReports,
+        lines: lineReports,
       });
 
       coverageSummary.totals.totalLineCount += lines.length;
@@ -264,6 +293,120 @@ class TestCoverage {
       }
     }
   }
+
+  getCoverageFromDirectory() {
+    const result = new SafeMap();
+    let dir;
+
+    try {
+      dir = opendirSync(this.coverageDirectory);
+
+      for (let entry; (entry = dir.readSync()) !== null;) {
+        if (RegExpPrototypeExec(kCoverageFileRegex, entry.name) === null) {
+          continue;
+        }
+
+        const coverageFile = join(this.coverageDirectory, entry.name);
+        const coverage = JSONParse(readFileSync(coverageFile, 'utf8'));
+        mergeCoverage(result, this.mapCoverageWithSourceMap(coverage));
+      }
+
+      return ArrayFrom(result.values());
+    } finally {
+      if (dir) {
+        dir.closeSync();
+      }
+    }
+  }
+
+
+  mapCoverageWithSourceMap(coverage) {
+    const { result } = coverage;
+    const sourceMapCache = coverage['source-map-cache'];
+    if (!sourceMapCache) {
+      return result;
+    }
+    const newResult = new SafeMap();
+    for (let i = 0; i < result.length; ++i) {
+      const script = result[i];
+      const { url, functions } = script;
+
+      if (shouldSkipFileCoverage(url) || sourceMapCache[url] == null) {
+        newResult.set(url, script);
+        continue;
+      }
+      const { data, lineLengths } = sourceMapCache[url];
+      let offset = 0;
+      const executedLines = ArrayPrototypeMap(lineLengths, (length, i) => {
+        const coverageLine = new CoverageLine(i + 1, offset, null, length);
+        offset += length;
+        return coverageLine;
+      });
+      if (data.sourcesContent != null) {
+        for (let j = 0; j < data.sources.length; ++j) {
+          this.getLines(data.sources[j], data.sourcesContent[j]);
+        }
+      }
+      const sourceMap = new SourceMap(data, { __proto__: null, lineLengths });
+
+      for (let j = 0; j < functions.length; ++j) {
+        const { ranges, functionName, isBlockCoverage } = functions[j];
+        if (ranges == null) {
+          continue;
+        }
+        let newUrl;
+        const newRanges = [];
+        for (let k = 0; k < ranges.length; ++k) {
+          const { startOffset, endOffset, count } = ranges[k];
+          const { lines } = mapRangeToLines(ranges[k], executedLines);
+
+          let startEntry = sourceMap
+            .findEntry(lines[0].line - 1, MathMax(0, startOffset - lines[0].startOffset));
+          const endEntry = sourceMap
+            .findEntry(lines[lines.length - 1].line - 1, (endOffset - lines[lines.length - 1].startOffset) - 1);
+          if (!startEntry.originalSource && endEntry.originalSource &&
+            lines[0].line === 1 && startOffset === 0 && lines[0].startOffset === 0) {
+            // Edge case when the first line is not mappable
+            const { 2: originalSource, 3: originalLine, 4: originalColumn } = sourceMap[kMappings][0];
+            startEntry = { __proto__: null, originalSource, originalLine, originalColumn };
+          }
+
+          if (!startEntry.originalSource || startEntry.originalSource !== endEntry.originalSource) {
+            // The range is not mappable. Skip it.
+            continue;
+          }
+
+          newUrl ??= startEntry?.originalSource;
+          const mappedLines = this.getLines(newUrl);
+          const mappedStartOffset = this.entryToOffset(startEntry, mappedLines);
+          const mappedEndOffset = this.entryToOffset(endEntry, mappedLines) + 1;
+          for (let l = startEntry.originalLine; l <= endEntry.originalLine; l++) {
+            mappedLines[l].count = count;
+          }
+
+          ArrayPrototypePush(newRanges, {
+            __proto__: null, startOffset: mappedStartOffset, endOffset: mappedEndOffset, count,
+          });
+        }
+
+        if (!newUrl) {
+          // No mappable ranges. Skip the function.
+          continue;
+        }
+        const newScript = newResult.get(newUrl) ?? { __proto__: null, url: newUrl, functions: [] };
+        ArrayPrototypePush(newScript.functions, { __proto__: null, functionName, ranges: newRanges, isBlockCoverage });
+        newResult.set(newUrl, newScript);
+      }
+    }
+
+    return ArrayFrom(newResult.values());
+  }
+
+  entryToOffset(entry, lines) {
+    const line = MathMax(entry.originalLine, 0);
+    return MathMin(lines[line].startOffset + entry.originalColumn, lines[line].endOffset);
+  }
+
 }
 
 function toPercentage(covered, total) {
@@ -317,9 +460,8 @@ function mapRangeToLines(range, lines) {
       while (endOffset > line?.startOffset) {
         // If the range is not covered, and the range covers the entire line,
         // then mark that line as not covered.
-        if (count === 0 && startOffset <= line.startOffset &&
-            endOffset >= line.endOffset) {
-          line.covered = false;
+        if (startOffset <= line.startOffset && endOffset >= line.endOffset) {
+          line.count = count;
         }
 
         ArrayPrototypePush(mappedLines, line);
@@ -340,36 +482,17 @@ function mapRangeToLines(range, lines) {
     }
   }
 
-  // Add some useful data to the range. The test runner has read these ranges
-  // from a file, so we own the data structures and can do what we want.
-  range.lines = mappedLines;
-  range.ignoredLines = ignoredLines;
+  return { __proto__: null, lines: mappedLines, ignoredLines };
 }
 
-function getCoverageFromDirectory(coverageDirectory) {
-  const result = new SafeMap();
-  let dir;
-
-  try {
-    dir = opendirSync(coverageDirectory);
-
-    for (let entry; (entry = dir.readSync()) !== null;) {
-      if (RegExpPrototypeExec(kCoverageFileRegex, entry.name) === null) {
-        continue;
-      }
-
-      const coverageFile = join(coverageDirectory, entry.name);
-      const coverage = JSONParse(readFileSync(coverageFile, 'utf8'));
-
-      mergeCoverage(result, coverage.result);
-    }
-
-    return ArrayFrom(result.values());
-  } finally {
-    if (dir) {
-      dir.closeSync();
-    }
-  }
+function shouldSkipFileCoverage(url) {
+  // The first part of this check filters out the node_modules/ directory
+  // from the results. This filter is applied first because most real world
+  // applications will be dominated by third party dependencies. The second
+  // part of the check filters out core modules, which start with 'node:' in
+  // coverage reports, as well as any invalid coverages which have been
+  // observed on Windows.
+  return StringPrototypeIncludes(url, '/node_modules/') || !StringPrototypeStartsWith(url, 'file:');
 }
 
 function mergeCoverage(merged, coverage) {
@@ -377,14 +500,7 @@ function mergeCoverage(merged, coverage) {
     const newScript = coverage[i];
     const { url } = newScript;
 
-    // The first part of this check filters out the node_modules/ directory
-    // from the results. This filter is applied first because most real world
-    // applications will be dominated by third party dependencies. The second
-    // part of the check filters out core modules, which start with 'node:' in
-    // coverage reports, as well as any invalid coverages which have been
-    // observed on Windows.
-    if (StringPrototypeIncludes(url, '/node_modules/') ||
-        !StringPrototypeStartsWith(url, 'file:')) {
+    if (shouldSkipFileCoverage(url)) {
       continue;
     }
 
