@@ -3,6 +3,7 @@ const {
   ArrayFrom,
   ArrayIsArray,
   ArrayPrototypeFilter,
+  ArrayPrototypeFind,
   ArrayPrototypeForEach,
   ArrayPrototypeIncludes,
   ArrayPrototypeMap,
@@ -41,7 +42,6 @@ const {
     ERR_INVALID_ARG_TYPE,
     ERR_INVALID_ARG_VALUE,
     ERR_TEST_FAILURE,
-    ERR_OUT_OF_RANGE,
   },
 } = require('internal/errors');
 const {
@@ -70,11 +70,13 @@ const {
   countCompletedTest,
   doesPathMatchFilter,
   isSupportedFileType,
+  shouldColorizeTestFiles,
 } = require('internal/test_runner/utils');
 const { basename, join, resolve } = require('path');
 const { once } = require('events');
 const {
   triggerUncaughtException,
+  exitCodes: { kGenericUserError },
 } = internalBinding('errors');
 
 const kFilterArgs = ['--test', '--experimental-test-coverage', '--watch'];
@@ -134,15 +136,12 @@ function createTestFileList() {
     for (let i = 0; i < testPaths.length; i++) {
       const absolutePath = resolve(testPaths[i]);
 
-      processPath(absolutePath, testFiles, {
-        __proto__: null,
-        userSupplied: true,
-      });
+      processPath(absolutePath, testFiles, { __proto__: null, userSupplied: true });
     }
   } catch (err) {
     if (err?.code === 'ENOENT') {
       console.error(`Could not find '${err.path}'`);
-      process.exit(1);
+      process.exit(kGenericUserError);
     }
 
     throw err;
@@ -156,8 +155,11 @@ function filterExecArgv(arg, i, arr) {
   !ArrayPrototypeSome(kFilterArgValues, (p) => arg === p || (i > 0 && arr[i - 1] === p) || StringPrototypeStartsWith(arg, `${p}=`));
 }
 
-function getRunArgs(path, { inspectPort, testNamePatterns, only }) {
+function getRunArgs(path, { forceExit, inspectPort, testNamePatterns, only }) {
   const argv = ArrayPrototypeFilter(process.execArgv, filterExecArgv);
+  if (forceExit === true) {
+    ArrayPrototypePush(argv, '--test-force-exit');
+  }
   if (isUsingInspector()) {
     ArrayPrototypePush(argv, `--inspect-port=${getInspectPort(inspectPort)}`);
   }
@@ -363,7 +365,7 @@ class FileTest extends Test {
 
 function runTestFile(path, filesWatcher, opts) {
   const watchMode = filesWatcher != null;
-  const subtest = opts.root.createSubtest(FileTest, path, async (t) => {
+  const subtest = opts.root.createSubtest(FileTest, path, { __proto__: null, signal: opts.signal }, async (t) => {
     const args = getRunArgs(path, opts);
     const stdio = ['pipe', 'pipe', 'pipe'];
     const env = { __proto__: null, ...process.env, NODE_TEST_CONTEXT: 'child-v8' };
@@ -417,9 +419,16 @@ function runTestFile(path, filesWatcher, opts) {
     if (watchMode) {
       filesWatcher.runningProcesses.delete(path);
       filesWatcher.runningSubtests.delete(path);
-      if (filesWatcher.runningSubtests.size === 0) {
-        opts.root.reporter[kEmitMessage]('test:watch:drained');
-      }
+      (async () => {
+        try {
+          await subTestEnded;
+        } finally {
+          if (filesWatcher.runningSubtests.size === 0) {
+            opts.root.reporter[kEmitMessage]('test:watch:drained');
+            opts.root.postRun();
+          }
+        }
+      })();
     }
 
     if (code !== 0 || signal !== null) {
@@ -438,7 +447,8 @@ function runTestFile(path, filesWatcher, opts) {
       throw err;
     }
   });
-  return subtest.start();
+  const subTestEnded = subtest.start();
+  return subTestEnded;
 }
 
 function watchFiles(testFiles, opts) {
@@ -446,8 +456,24 @@ function watchFiles(testFiles, opts) {
   const runningSubtests = new SafeMap();
   const watcher = new FilesWatcher({ __proto__: null, debounce: 200, mode: 'filter', signal: opts.signal });
   const filesWatcher = { __proto__: null, watcher, runningProcesses, runningSubtests };
+  opts.root.harness.watching = true;
+
+  watcher.on('changed', ({ owners, eventType }) => {
+    if (eventType === 'rename') {
+      const updatedTestFiles = createTestFileList();
+
+      const newFileName = ArrayPrototypeFind(updatedTestFiles, (x) => !ArrayPrototypeIncludes(testFiles, x));
+      const previousFileName = ArrayPrototypeFind(testFiles, (x) => !ArrayPrototypeIncludes(updatedTestFiles, x));
+
+      // When file renamed
+      if (newFileName && previousFileName) {
+        owners = new SafeSet().add(newFileName);
+        watcher.filterFile(resolve(newFileName), owners);
+      }
+
+      testFiles = updatedTestFiles;
+    }
 
-  watcher.on('changed', ({ owners }) => {
     watcher.unfilterFilesOwnedBy(owners);
     PromisePrototypeThen(SafePromiseAllReturnVoid(testFiles, async (file) => {
       if (!owners.has(file)) {
@@ -472,7 +498,10 @@ function watchFiles(testFiles, opts) {
     kResistStopPropagation ??= require('internal/event_target').kResistStopPropagation;
     opts.signal.addEventListener(
       'abort',
-      () => opts.root.postRun(),
+      () => {
+        opts.root.harness.watching = false;
+        opts.root.postRun();
+      },
       { __proto__: null, once: true, [kResistStopPropagation]: true },
     );
   }
@@ -485,7 +514,18 @@ function run(options) {
     options = kEmptyObject;
   }
   let { testNamePatterns, shard } = options;
-  const { concurrency, timeout, signal, files, inspectPort, watch, setup, only } = options;
+  const {
+    concurrency,
+    timeout,
+    signal,
+    files,
+    forceExit,
+    inspectPort,
+    watch,
+    setup,
+    only,
+    plan,
+  } = options;
 
   if (files != null) {
     validateArray(files, 'options.files');
@@ -493,6 +533,15 @@ function run(options) {
   if (watch != null) {
     validateBoolean(watch, 'options.watch');
   }
+  if (forceExit != null) {
+    validateBoolean(forceExit, 'options.forceExit');
+
+    if (forceExit && watch) {
+      throw new ERR_INVALID_ARG_VALUE(
+        'options.forceExit', watch, 'is not supported with watch mode',
+      );
+    }
+  }
   if (only != null) {
     validateBoolean(only, 'options.only');
   }
@@ -502,11 +551,7 @@ function run(options) {
     shard = { __proto__: null, index: shard.index, total: shard.total };
 
     validateInteger(shard.total, 'options.shard.total', 1);
-    validateInteger(shard.index, 'options.shard.index');
-
-    if (shard.index <= 0 || shard.total < shard.index) {
-      throw new ERR_OUT_OF_RANGE('options.shard.index', `>= 1 && <= ${shard.total} ("options.shard.total")`, shard.index);
-    }
+    validateInteger(shard.index, 'options.shard.index', 1, shard.total);
 
     if (watch) {
       throw new ERR_INVALID_ARG_VALUE('options.shard', watch, 'shards not supported with watch mode');
@@ -532,7 +577,14 @@ function run(options) {
     });
   }
 
-  const root = createTestTree({ __proto__: null, concurrency, timeout, signal });
+  const root = createTestTree({ __proto__: null, concurrency, timeout, signal, plan });
+  root.harness.shouldColorizeTestFiles ||= shouldColorizeTestFiles(root);
+
+  if (process.env.NODE_TEST_CONTEXT !== undefined) {
+    process.emitWarning('node:test run() is being called recursively within a test file. skipping running files.');
+    root.postRun();
+    return root.reporter;
+  }
   let testFiles = files ?? createTestFileList();
 
   if (shard) {
@@ -541,13 +593,22 @@ function run(options) {
 
   let postRun = () => root.postRun();
   let filesWatcher;
-  const opts = { __proto__: null, root, signal, inspectPort, testNamePatterns, only };
+  const opts = {
+    __proto__: null,
+    root,
+    signal,
+    inspectPort,
+    testNamePatterns,
+    only,
+    forceExit,
+  };
   if (watch) {
     filesWatcher = watchFiles(testFiles, opts);
     postRun = undefined;
   }
   const runFiles = () => {
     root.harness.bootstrapComplete = true;
+    root.harness.allowTestsToRun = true;
     return SafePromiseAllSettledReturnVoid(testFiles, (path) => {
       const subtest = runTestFile(path, filesWatcher, opts);
       filesWatcher?.runningSubtests.set(path, subtest);
@@ -555,7 +616,7 @@ function run(options) {
     });
   };
 
-  PromisePrototypeThen(PromisePrototypeThen(PromiseResolve(setup?.(root)), runFiles), postRun);
+  PromisePrototypeThen(PromisePrototypeThen(PromiseResolve(setup?.(root.reporter)), runFiles), postRun);
 
   return root.reporter;
 }
