@@ -11,7 +11,6 @@ const {
   FunctionPrototype,
   MathMax,
   Number,
-  ObjectEntries,
   ObjectSeal,
   PromisePrototypeThen,
   PromiseResolve,
@@ -103,10 +102,28 @@ function lazyAssertObject() {
   if (assertObj === undefined) {
     assertObj = new SafeMap();
     const assert = require('assert');
-    for (const { 0: key, 1: value } of ObjectEntries(assert)) {
-      if (typeof value === 'function') {
-        assertObj.set(value, key);
-      }
+
+    const methodsToCopy = [
+      'deepEqual',
+      'deepStrictEqual',
+      'doesNotMatch',
+      'doesNotReject',
+      'doesNotThrow',
+      'equal',
+      'fail',
+      'ifError',
+      'match',
+      'notDeepEqual',
+      'notDeepStrictEqual',
+      'notEqual',
+      'notStrictEqual',
+      'ok',
+      'rejects',
+      'strictEqual',
+      'throws',
+    ];
+    for (let i = 0; i < methodsToCopy.length; i++) {
+      assertObj.set(methodsToCopy[i], assert[methodsToCopy[i]]);
     }
   }
   return assertObj;
@@ -153,7 +170,7 @@ function stopTest(timeout, signal) {
 
 class TestPlan {
   constructor(count) {
-    validateUint32(count, 'count', 0);
+    validateUint32(count, 'count');
     this.expected = count;
     this.actual = 0;
   }
@@ -184,6 +201,10 @@ class TestContext {
     return this.#test.name;
   }
 
+  get fullName() {
+    return getFullName(this.#test);
+  }
+
   get error() {
     return this.#test.error;
   }
@@ -210,18 +231,18 @@ class TestContext {
   get assert() {
     if (this.#assert === undefined) {
       const { plan } = this.#test;
-      const assertions = lazyAssertObject();
+      const map = lazyAssertObject();
       const assert = { __proto__: null };
 
       this.#assert = assert;
-      for (const { 0: method, 1: name } of assertions.entries()) {
+      map.forEach((method, name) => {
         assert[name] = (...args) => {
           if (plan !== null) {
             plan.actual++;
           }
           return ReflectApply(method, assert, args);
         };
-      }
+      });
     }
     return this.#assert;
   }
@@ -297,6 +318,10 @@ class SuiteContext {
   get name() {
     return this.#suite.name;
   }
+
+  get fullName() {
+    return getFullName(this.#suite);
+  }
 }
 
 class Test extends AsyncResource {
@@ -363,7 +388,7 @@ class Test extends AsyncResource {
 
     switch (typeof concurrency) {
       case 'number':
-        validateUint32(concurrency, 'options.concurrency', 1);
+        validateUint32(concurrency, 'options.concurrency', true);
         this.concurrency = concurrency;
         break;
 
@@ -1122,6 +1147,16 @@ class Suite extends Test {
   }
 }
 
+function getFullName(test) {
+  let fullName = test.name;
+
+  for (let t = test.parent; t !== t.root; t = t.parent) {
+    fullName = `${t.name} > ${fullName}`;
+  }
+
+  return fullName;
+}
+
 module.exports = {
   kCancelledByParent,
   kSubtestsFailed,
