@@ -11,7 +11,6 @@ const {
   FunctionPrototypeCall,
   MathMin,
   NumberIsInteger,
-  ObjectCreate,
   ObjectDefineProperties,
   ObjectSetPrototypeOf,
   Promise,
@@ -60,6 +59,8 @@ const {
   validateAbortSignal,
   validateBuffer,
   validateObject,
+  kValidateObjectAllowNullable,
+  kValidateObjectAllowFunction,
 } = require('internal/validators');
 
 const {
@@ -85,9 +86,7 @@ const {
   kControllerErrorFunction,
 } = require('internal/streams/utils');
 
-const {
-  structuredClone,
-} = require('internal/structured_clone');
+const { structuredClone } = internalBinding('messaging');
 
 const {
   ArrayBufferViewGetBuffer,
@@ -96,9 +95,9 @@ const {
   AsyncIterator,
   cloneAsUint8Array,
   copyArrayBuffer,
+  createPromiseCallback,
   customInspect,
   dequeueValue,
-  ensureIsPromise,
   enqueueValueWithSize,
   extractHighWaterMark,
   extractSizeAlgorithm,
@@ -111,6 +110,8 @@ const {
   nonOpCancel,
   nonOpPull,
   nonOpStart,
+  getIterator,
+  iteratorNext,
   kType,
   kState,
 } = require('internal/webstreams/util');
@@ -249,19 +250,7 @@ class ReadableStream {
   constructor(source = {}, strategy = kEmptyObject) {
     if (source === null)
       throw new ERR_INVALID_ARG_VALUE('source', 'Object', source);
-    this[kState] = {
-      disturbed: false,
-      reader: undefined,
-      state: 'readable',
-      storedError: undefined,
-      stream: undefined,
-      transfer: {
-        writable: undefined,
-        port1: undefined,
-        port2: undefined,
-        promise: undefined,
-      },
-    };
+    this[kState] = createReadableStreamState();
 
     this[kIsClosedPromise] = createDeferredPromise();
     this[kControllerErrorFunction] = () => {};
@@ -319,6 +308,10 @@ class ReadableStream {
     return isReadableStreamLocked(this);
   }
 
+  static from(iterable) {
+    return readableStreamFromIterable(iterable);
+  }
+
   /**
    * @param {any} [reason]
    * @returns { Promise<void> }
@@ -342,7 +335,7 @@ class ReadableStream {
   getReader(options = kEmptyObject) {
     if (!isReadableStream(this))
       throw new ERR_INVALID_THIS('ReadableStream');
-    validateObject(options, 'options', { nullable: true, allowFunction: true });
+    validateObject(options, 'options', kValidateObjectAllowNullable | kValidateObjectAllowFunction);
     const mode = options?.mode;
 
     if (mode === undefined)
@@ -473,9 +466,13 @@ class ReadableStream {
 
     // eslint-disable-next-line no-use-before-define
     const reader = new ReadableStreamDefaultReader(this);
-    let done = false;
+
+    // No __proto__ here to avoid the performance hit.
+    const state = {
+      done: false,
+      current: undefined,
+    };
     let started = false;
-    let current;
 
     // The nextSteps function is not an async function in order
     // to make it more efficient. Because nextSteps explicitly
@@ -484,7 +481,7 @@ class ReadableStream {
     // unnecessary Promise allocations to occur, which just add
     // cost.
     function nextSteps() {
-      if (done)
+      if (state.done)
         return PromiseResolve({ done: true, value: undefined });
 
       if (reader[kState].stream === undefined) {
@@ -494,31 +491,15 @@ class ReadableStream {
       }
       const promise = createDeferredPromise();
 
-      readableStreamDefaultReaderRead(reader, {
-        [kChunk](chunk) {
-          current = undefined;
-          promise.resolve({ value: chunk, done: false });
-        },
-        [kClose]() {
-          current = undefined;
-          done = true;
-          readableStreamReaderGenericRelease(reader);
-          promise.resolve({ done: true, value: undefined });
-        },
-        [kError](error) {
-          current = undefined;
-          done = true;
-          readableStreamReaderGenericRelease(reader);
-          promise.reject(error);
-        },
-      });
+      // eslint-disable-next-line no-use-before-define
+      readableStreamDefaultReaderRead(reader, new ReadableStreamAsyncIteratorReadRequest(reader, state, promise));
       return promise.promise;
     }
 
     async function returnSteps(value) {
-      if (done)
-        return { done: true, value };
-      done = true;
+      if (state.done)
+        return { done: true, value }; // eslint-disable-line node-core/avoid-prototype-pollution
+      state.done = true;
 
       if (reader[kState].stream === undefined) {
         throw new ERR_INVALID_STATE.TypeError(
@@ -529,11 +510,11 @@ class ReadableStream {
         const result = readableStreamReaderGenericCancel(reader, value);
         readableStreamReaderGenericRelease(reader);
         await result;
-        return { done: true, value };
+        return { done: true, value }; // eslint-disable-line node-core/avoid-prototype-pollution
       }
 
       readableStreamReaderGenericRelease(reader);
-      return { done: true, value };
+      return { done: true, value }; // eslint-disable-line node-core/avoid-prototype-pollution
     }
 
     // TODO(@jasnell): Explore whether an async generator
@@ -555,19 +536,19 @@ class ReadableStream {
         // need to investigate if it's a bug in our impl or
         // the spec.
         if (!started) {
-          current = PromiseResolve();
+          state.current = PromiseResolve();
           started = true;
         }
-        current = current !== undefined ?
-          PromisePrototypeThen(current, nextSteps, nextSteps) :
+        state.current = state.current !== undefined ?
+          PromisePrototypeThen(state.current, nextSteps, nextSteps) :
           nextSteps();
-        return current;
+        return state.current;
       },
 
       return(error) {
-        return current ?
+        return state.current ?
           PromisePrototypeThen(
-            current,
+            state.current,
             () => returnSteps(error),
             () => returnSteps(error)) :
           returnSteps(error);
@@ -627,7 +608,11 @@ class ReadableStream {
     const transfer = lazyTransfer();
     setupReadableStreamDefaultControllerFromSource(
       this,
-      new transfer.CrossRealmTransformReadableSource(port),
+      // The MessagePort is set to be referenced when reading.
+      // After two MessagePorts are closed, there is a problem with
+      // lingering promise not being properly resolved.
+      // https://github.com/nodejs/node/issues/51486
+      new transfer.CrossRealmTransformReadableSource(port, true),
       0, () => 1);
   }
 }
@@ -646,24 +631,18 @@ ObjectDefineProperties(ReadableStream.prototype, {
   pipeThrough: kEnumerableProperty,
   pipeTo: kEnumerableProperty,
   tee: kEnumerableProperty,
+  values: kEnumerableProperty,
   [SymbolToStringTag]: getNonWritablePropertyDescriptor(ReadableStream.name),
 });
+ObjectDefineProperties(ReadableStream, {
+  from: kEnumerableProperty,
+});
 
 function TransferredReadableStream() {
   return makeTransferable(ReflectConstruct(
     function() {
       this[kType] = 'ReadableStream';
-      this[kState] = {
-        disturbed: false,
-        state: 'readable',
-        storedError: undefined,
-        stream: undefined,
-        transfer: {
-          writable: undefined,
-          port: undefined,
-          promise: undefined,
-        },
-      };
+      this[kState] = createReadableStreamState();
       this[kIsClosedPromise] = createDeferredPromise();
     },
     [], ReadableStream));
@@ -768,6 +747,33 @@ function createReadableStreamBYOBRequest(controller, view) {
   return stream;
 }
 
+class ReadableStreamAsyncIteratorReadRequest {
+  constructor(reader, state, promise) {
+    this.reader = reader;
+    this.state = state;
+    this.promise = promise;
+  }
+
+  [kChunk](chunk) {
+    this.state.current = undefined;
+    this.promise.resolve({ value: chunk, done: false });
+  }
+
+  [kClose]() {
+    this.state.current = undefined;
+    this.state.done = true;
+    readableStreamReaderGenericRelease(this.reader);
+    this.promise.resolve({ done: true, value: undefined });
+  }
+
+  [kError](error) {
+    this.state.current = undefined;
+    this.state.done = true;
+    readableStreamReaderGenericRelease(this.reader);
+    this.promise.reject(error);
+  }
+}
+
 class DefaultReadRequest {
   constructor() {
     this[kState] = createDeferredPromise();
@@ -1194,34 +1200,58 @@ ObjectDefineProperties(ReadableByteStreamController.prototype, {
   [SymbolToStringTag]: getNonWritablePropertyDescriptor(ReadableByteStreamController.name),
 });
 
-function createTeeReadableStream(start, pull, cancel) {
-  return ReflectConstruct(
-    function() {
-      this[kType] = 'ReadableStream';
-      this[kState] = {
-        disturbed: false,
-        state: 'readable',
-        storedError: undefined,
-        stream: undefined,
-        transfer: {
-          writable: undefined,
-          port: undefined,
-          promise: undefined,
-        },
-      };
-      this[kIsClosedPromise] = createDeferredPromise();
-      setupReadableStreamDefaultControllerFromSource(
-        this,
-        ObjectCreate(null, {
-          start: { __proto__: null, value: start },
-          pull: { __proto__: null, value: pull },
-          cancel: { __proto__: null, value: cancel },
-        }),
-        1,
-        () => 1);
-      return makeTransferable(this);
-    }, [], ReadableStream,
-  );
+function InternalReadableStream(start, pull, cancel, highWaterMark, size) {
+  this[kType] = 'ReadableStream';
+  this[kState] = createReadableStreamState();
+  this[kIsClosedPromise] = createDeferredPromise();
+  const controller = new ReadableStreamDefaultController(kSkipThrow);
+  setupReadableStreamDefaultController(
+    this,
+    controller,
+    start,
+    pull,
+    cancel,
+    highWaterMark,
+    size);
+  return makeTransferable(this);
+}
+
+ObjectSetPrototypeOf(InternalReadableStream.prototype, ReadableStream.prototype);
+ObjectSetPrototypeOf(InternalReadableStream, ReadableStream);
+
+function createReadableStream(start, pull, cancel, highWaterMark = 1, size = () => 1) {
+  const stream = new InternalReadableStream(start, pull, cancel, highWaterMark, size);
+
+  // For spec compliance the InternalReadableStream must be a ReadableStream
+  stream.constructor = ReadableStream;
+  return stream;
+}
+
+function InternalReadableByteStream(start, pull, cancel) {
+  this[kType] = 'ReadableStream';
+  this[kState] = createReadableStreamState();
+  this[kIsClosedPromise] = createDeferredPromise();
+  const controller = new ReadableByteStreamController(kSkipThrow);
+  setupReadableByteStreamController(
+    this,
+    controller,
+    start,
+    pull,
+    cancel,
+    0,
+    undefined);
+  return makeTransferable(this);
+}
+
+ObjectSetPrototypeOf(InternalReadableByteStream.prototype, ReadableStream.prototype);
+ObjectSetPrototypeOf(InternalReadableByteStream, ReadableStream);
+
+function createReadableByteStream(start, pull, cancel) {
+  const stream = new InternalReadableByteStream(start, pull, cancel);
+
+  // For spec compliance the InternalReadableByteStream must be a ReadableStream
+  stream.constructor = ReadableStream;
+  return stream;
 }
 
 const isReadableStream =
@@ -1237,6 +1267,72 @@ const isReadableStreamBYOBReader =
 
 // ---- ReadableStream Implementation
 
+function createReadableStreamState() {
+  return {
+    __proto__: null,
+    disturbed: false,
+    reader: undefined,
+    state: 'readable',
+    storedError: undefined,
+    transfer: {
+      __proto__: null,
+      writable: undefined,
+      port1: undefined,
+      port2: undefined,
+      promise: undefined,
+    },
+  };
+}
+
+function readableStreamFromIterable(iterable) {
+  let stream;
+  const iteratorRecord = getIterator(iterable, 'async');
+
+  const startAlgorithm = nonOpStart;
+
+  async function pullAlgorithm() {
+    const nextResult = iteratorNext(iteratorRecord);
+    const nextPromise = PromiseResolve(nextResult);
+    return PromisePrototypeThen(nextPromise, (iterResult) => {
+      if (typeof iterResult !== 'object' || iterResult === null) {
+        throw new ERR_INVALID_STATE.TypeError(
+          'The promise returned by the iterator.next() method must fulfill with an object');
+      }
+      if (iterResult.done) {
+        readableStreamDefaultControllerClose(stream[kState].controller);
+      } else {
+        readableStreamDefaultControllerEnqueue(stream[kState].controller, iterResult.value);
+      }
+    });
+  }
+
+  async function cancelAlgorithm(reason) {
+    const iterator = iteratorRecord.iterator;
+    const returnMethod = iterator.return;
+    if (returnMethod === undefined) {
+      return PromiseResolve();
+    }
+    const returnResult = FunctionPrototypeCall(returnMethod, iterator, reason);
+    const returnPromise = PromiseResolve(returnResult);
+    return PromisePrototypeThen(returnPromise, (iterResult) => {
+      if (typeof iterResult !== 'object' || iterResult === null) {
+        throw new ERR_INVALID_STATE.TypeError(
+          'The promise returned by the iterator.return() method must fulfill with an object');
+      }
+      return undefined;
+    });
+  }
+
+  stream = createReadableStream(
+    startAlgorithm,
+    pullAlgorithm,
+    cancelAlgorithm,
+    0,
+  );
+
+  return stream;
+}
+
 function readableStreamPipeTo(
   source,
   dest,
@@ -1271,7 +1367,9 @@ function readableStreamPipeTo(
 
   const promise = createDeferredPromise();
 
-  let currentWrite = PromiseResolve();
+  const state = {
+    currentWrite: PromiseResolve(),
+  };
 
   // The error here can be undefined. The rejected arg
   // tells us that the promise must be rejected even
@@ -1288,9 +1386,9 @@ function readableStreamPipeTo(
   }
 
   async function waitForCurrentWrite() {
-    const write = currentWrite;
+    const write = state.currentWrite;
     await write;
-    if (write !== currentWrite)
+    if (write !== state.currentWrite)
       await waitForCurrentWrite();
   }
 
@@ -1381,20 +1479,14 @@ function readableStreamPipeTo(
   async function step() {
     if (shuttingDown)
       return true;
+
     await writer[kState].ready.promise;
-    return new Promise((resolve, reject) => {
-      readableStreamDefaultReaderRead(
-        reader,
-        {
-          [kChunk](chunk) {
-            currentWrite = writableStreamDefaultWriterWrite(writer, chunk);
-            setPromiseHandled(currentWrite);
-            resolve(false);
-          },
-          [kClose]: () => resolve(true),
-          [kError]: reject,
-        });
-    });
+
+    const promise = createDeferredPromise();
+    // eslint-disable-next-line no-use-before-define
+    readableStreamDefaultReaderRead(reader, new PipeToReadableStreamReadRequest(writer, state, promise));
+
+    return promise.promise;
   }
 
   async function run() {
@@ -1407,7 +1499,7 @@ function readableStreamPipeTo(
       abortAlgorithm();
       return promise.promise;
     }
-    addAbortListener ??= require('events').addAbortListener;
+    addAbortListener ??= require('internal/events/abort_listener').addAbortListener;
     disposable = addAbortListener(signal, abortAlgorithm);
   }
 
@@ -1456,6 +1548,28 @@ function readableStreamPipeTo(
   return promise.promise;
 }
 
+class PipeToReadableStreamReadRequest {
+  constructor(writer, state, promise) {
+    this.writer = writer;
+    this.state = state;
+    this.promise = promise;
+  }
+
+  [kChunk](chunk) {
+    this.state.currentWrite = writableStreamDefaultWriterWrite(this.writer, chunk);
+    setPromiseHandled(this.state.currentWrite);
+    this.promise.resolve(false);
+  }
+
+  [kClose]() {
+    this.promise.resolve(true);
+  }
+
+  [kError](error) {
+    this.promise.reject(error);
+  }
+}
+
 function readableStreamTee(stream, cloneForBranch2) {
   if (isReadableByteStreamController(stream[kState].controller)) {
     return readableByteStreamTee(stream);
@@ -1540,9 +1654,9 @@ function readableStreamDefaultTee(stream, cloneForBranch2) {
   }
 
   branch1 =
-    createTeeReadableStream(nonOpStart, pullAlgorithm, cancel1Algorithm);
+    createReadableStream(nonOpStart, pullAlgorithm, cancel1Algorithm);
   branch2 =
-    createTeeReadableStream(nonOpStart, pullAlgorithm, cancel2Algorithm);
+    createReadableStream(nonOpStart, pullAlgorithm, cancel2Algorithm);
 
   PromisePrototypeThen(
     reader[kState].close.promise,
@@ -1819,16 +1933,10 @@ function readableByteStreamTee(stream) {
     return cancelDeferred.promise;
   }
 
-  branch1 = new ReadableStream({
-    type: 'bytes',
-    pull: pull1Algorithm,
-    cancel: cancel1Algorithm,
-  });
-  branch2 = new ReadableStream({
-    type: 'bytes',
-    pull: pull2Algorithm,
-    cancel: cancel2Algorithm,
-  });
+  branch1 =
+    createReadableByteStream(nonOpStart, pull1Algorithm, cancel1Algorithm);
+  branch2 =
+    createReadableByteStream(nonOpStart, pull2Algorithm, cancel2Algorithm);
 
   forwardReaderError(reader);
 
@@ -1879,10 +1987,7 @@ function readableStreamCancel(stream, reason) {
   }
 
   return PromisePrototypeThen(
-    ensureIsPromise(
-      stream[kState].controller[kCancel],
-      stream[kState].controller,
-      reason),
+    stream[kState].controller[kCancel](reason),
     () => {});
 }
 
@@ -2247,7 +2352,7 @@ function readableStreamDefaultControllerCallPullIfNeeded(controller) {
   assert(!controller[kState].pullAgain);
   controller[kState].pulling = true;
   PromisePrototypeThen(
-    ensureIsPromise(controller[kState].pullAlgorithm, controller),
+    controller[kState].pullAlgorithm(controller),
     () => {
       controller[kState].pulling = false;
       if (controller[kState].pullAgain) {
@@ -2277,12 +2382,9 @@ function readableStreamDefaultControllerError(controller, error) {
 
 function readableStreamDefaultControllerCancelSteps(controller, reason) {
   resetQueue(controller);
-  try {
-    const result = controller[kState].cancelAlgorithm(reason);
-    return result;
-  } finally {
-    readableStreamDefaultControllerClearAlgorithms(controller);
-  }
+  const result = controller[kState].cancelAlgorithm(reason);
+  readableStreamDefaultControllerClearAlgorithms(controller);
+  return result;
 }
 
 function readableStreamDefaultControllerPullSteps(controller, readRequest) {
@@ -2333,7 +2435,7 @@ function setupReadableStreamDefaultController(
   const startResult = startAlgorithm();
 
   PromisePrototypeThen(
-    PromiseResolve(startResult),
+    new Promise((r) => r(startResult)),
     () => {
       controller[kState].started = true;
       assert(!controller[kState].pulling);
@@ -2356,11 +2458,10 @@ function setupReadableStreamDefaultControllerFromSource(
     FunctionPrototypeBind(start, source, controller) :
     nonOpStart;
   const pullAlgorithm = pull ?
-    FunctionPrototypeBind(pull, source, controller) :
+    createPromiseCallback('source.pull', pull, source) :
     nonOpPull;
-
   const cancelAlgorithm = cancel ?
-    FunctionPrototypeBind(cancel, source) :
+    createPromiseCallback('source.cancel', cancel, source) :
     nonOpCancel;
 
   setupReadableStreamDefaultController(
@@ -2988,7 +3089,7 @@ function readableByteStreamControllerCallPullIfNeeded(controller) {
   assert(!controller[kState].pullAgain);
   controller[kState].pulling = true;
   PromisePrototypeThen(
-    ensureIsPromise(controller[kState].pullAlgorithm, controller),
+    controller[kState].pullAlgorithm(controller),
     () => {
       controller[kState].pulling = false;
       if (controller[kState].pullAgain) {
@@ -3132,7 +3233,7 @@ function setupReadableByteStreamController(
   const startResult = startAlgorithm();
 
   PromisePrototypeThen(
-    PromiseResolve(startResult),
+    new Promise((r) => r(startResult)),
     () => {
       controller[kState].started = true;
       assert(!controller[kState].pulling);
@@ -3155,10 +3256,10 @@ function setupReadableByteStreamControllerFromSource(
     FunctionPrototypeBind(start, source, controller) :
     nonOpStart;
   const pullAlgorithm = pull ?
-    FunctionPrototypeBind(pull, source, controller) :
+    createPromiseCallback('source.pull', pull, source, controller) :
     nonOpPull;
   const cancelAlgorithm = cancel ?
-    FunctionPrototypeBind(cancel, source) :
+    createPromiseCallback('source.cancel', cancel, source) :
     nonOpCancel;
 
   if (autoAllocateChunkSize === 0) {
@@ -3255,4 +3356,6 @@ module.exports = {
   readableByteStreamControllerPullSteps,
   setupReadableByteStreamController,
   setupReadableByteStreamControllerFromSource,
+  createReadableStream,
+  createReadableByteStream,
 };
