@@ -6,6 +6,7 @@ const {
   FunctionPrototypeBind,
   FunctionPrototypeCall,
   ObjectDefineProperties,
+  ObjectSetPrototypeOf,
   Promise,
   PromisePrototypeThen,
   PromiseResolve,
@@ -49,9 +50,9 @@ const {
 } = require('internal/worker/js_transferable');
 
 const {
+  createPromiseCallback,
   customInspect,
   dequeueValue,
-  ensureIsPromise,
   enqueueValueWithSize,
   extractHighWaterMark,
   extractSizeAlgorithm,
@@ -159,45 +160,7 @@ class WritableStream {
     if (type !== undefined)
       throw new ERR_INVALID_ARG_VALUE.RangeError('type', type);
 
-    this[kState] = {
-      close: createDeferredPromise(),
-      closeRequest: {
-        promise: undefined,
-        resolve: undefined,
-        reject: undefined,
-      },
-      inFlightWriteRequest: {
-        promise: undefined,
-        resolve: undefined,
-        reject: undefined,
-      },
-      inFlightCloseRequest: {
-        promise: undefined,
-        resolve: undefined,
-        reject: undefined,
-      },
-      pendingAbortRequest: {
-        abort: {
-          promise: undefined,
-          resolve: undefined,
-          reject: undefined,
-        },
-        reason: undefined,
-        wasAlreadyErroring: false,
-      },
-      backpressure: false,
-      controller: undefined,
-      state: 'writable',
-      storedError: undefined,
-      writeRequests: [],
-      writer: undefined,
-      transfer: {
-        readable: undefined,
-        port1: undefined,
-        port2: undefined,
-        promise: undefined,
-      },
-    };
+    this[kState] = createWritableStreamState();
 
     this[kIsClosedPromise] = createDeferredPromise();
     this[kControllerErrorFunction] = () => {};
@@ -335,45 +298,7 @@ function TransferredWritableStream() {
   return makeTransferable(ReflectConstruct(
     function() {
       this[kType] = 'WritableStream';
-      this[kState] = {
-        close: createDeferredPromise(),
-        closeRequest: {
-          promise: undefined,
-          resolve: undefined,
-          reject: undefined,
-        },
-        inFlightWriteRequest: {
-          promise: undefined,
-          resolve: undefined,
-          reject: undefined,
-        },
-        inFlightCloseRequest: {
-          promise: undefined,
-          resolve: undefined,
-          reject: undefined,
-        },
-        pendingAbortRequest: {
-          abort: {
-            promise: undefined,
-            resolve: undefined,
-            reject: undefined,
-          },
-          reason: undefined,
-          wasAlreadyErroring: false,
-        },
-        backpressure: false,
-        controller: undefined,
-        state: 'writable',
-        storedError: undefined,
-        writeRequests: [],
-        writer: undefined,
-        transfer: {
-          promise: undefined,
-          port1: undefined,
-          port2: undefined,
-          readable: undefined,
-        },
-      };
+      this[kState] = createWritableStreamState();
       this[kIsClosedPromise] = createDeferredPromise();
       this[kControllerErrorFunction] = () => {};
     },
@@ -576,6 +501,35 @@ ObjectDefineProperties(WritableStreamDefaultController.prototype, {
   [SymbolToStringTag]: getNonWritablePropertyDescriptor(WritableStreamDefaultController.name),
 });
 
+function InternalWritableStream(start, write, close, abort, highWaterMark, size) {
+  this[kType] = 'WritableStream';
+  this[kState] = createWritableStreamState();
+  this[kIsClosedPromise] = createDeferredPromise();
+  const controller = new WritableStreamDefaultController(kSkipThrow);
+  setupWritableStreamDefaultController(
+    this,
+    controller,
+    start,
+    write,
+    close,
+    abort,
+    highWaterMark,
+    size,
+  );
+  return makeTransferable(this);
+}
+
+ObjectSetPrototypeOf(InternalWritableStream.prototype, WritableStream.prototype);
+ObjectSetPrototypeOf(InternalWritableStream, WritableStream);
+
+function createWritableStream(start, write, close, abort, highWaterMark = 1, size = () => 1) {
+  const stream = new InternalWritableStream(start, write, close, abort, highWaterMark, size);
+
+  // For spec compliance the InternalWritableStream must be a WritableStream
+  stream.constructor = WritableStream;
+  return stream;
+}
+
 const isWritableStream =
   isBrandCheck('WritableStream');
 const isWritableStreamDefaultWriter =
@@ -583,6 +537,55 @@ const isWritableStreamDefaultWriter =
 const isWritableStreamDefaultController =
   isBrandCheck('WritableStreamDefaultController');
 
+function createWritableStreamState() {
+  return {
+    __proto__: null,
+    close: createDeferredPromise(),
+    closeRequest: {
+      __proto__: null,
+      promise: undefined,
+      resolve: undefined,
+      reject: undefined,
+    },
+    inFlightWriteRequest: {
+      __proto__: null,
+      promise: undefined,
+      resolve: undefined,
+      reject: undefined,
+    },
+    inFlightCloseRequest: {
+      __proto__: null,
+      promise: undefined,
+      resolve: undefined,
+      reject: undefined,
+    },
+    pendingAbortRequest: {
+      __proto__: null,
+      abort: {
+        __proto__: null,
+        promise: undefined,
+        resolve: undefined,
+        reject: undefined,
+      },
+      reason: undefined,
+      wasAlreadyErroring: false,
+    },
+    backpressure: false,
+    controller: undefined,
+    state: 'writable',
+    storedError: undefined,
+    writeRequests: [],
+    writer: undefined,
+    transfer: {
+      __proto__: null,
+      readable: undefined,
+      port1: undefined,
+      port2: undefined,
+      promise: undefined,
+    },
+  };
+}
+
 function isWritableStreamLocked(stream) {
   return stream[kState].writer !== undefined;
 }
@@ -902,10 +905,7 @@ function writableStreamFinishErroring(stream) {
     return;
   }
   PromisePrototypeThen(
-    ensureIsPromise(
-      stream[kState].controller[kAbort],
-      stream[kState].controller,
-      abortRequest.reason),
+    stream[kState].controller[kAbort](abortRequest.reason),
     () => {
       abortRequest.abort.resolve?.();
       writableStreamRejectCloseAndClosedPromiseIfNeeded(stream);
@@ -1109,7 +1109,7 @@ function writableStreamDefaultControllerProcessWrite(controller, chunk) {
   writableStreamMarkFirstWriteRequestInFlight(stream);
 
   PromisePrototypeThen(
-    ensureIsPromise(writeAlgorithm, controller, chunk, controller),
+    writeAlgorithm(chunk, controller),
     () => {
       writableStreamFinishInFlightWrite(stream);
       const {
@@ -1142,7 +1142,7 @@ function writableStreamDefaultControllerProcessClose(controller) {
   writableStreamMarkCloseRequestInFlight(stream);
   dequeueValue(controller);
   assert(!queue.length);
-  const sinkClosePromise = ensureIsPromise(closeAlgorithm, controller);
+  const sinkClosePromise = closeAlgorithm();
   writableStreamDefaultControllerClearAlgorithms(controller);
   PromisePrototypeThen(
     sinkClosePromise,
@@ -1241,12 +1241,14 @@ function setupWritableStreamDefaultControllerFromSink(
     FunctionPrototypeBind(start, sink, controller) :
     nonOpStart;
   const writeAlgorithm = write ?
-    FunctionPrototypeBind(write, sink) :
+    createPromiseCallback('sink.write', write, sink) :
     nonOpWrite;
   const closeAlgorithm = close ?
-    FunctionPrototypeBind(close, sink) : nonOpCancel;
+    createPromiseCallback('sink.close', close, sink) :
+    nonOpCancel;
   const abortAlgorithm = abort ?
-    FunctionPrototypeBind(abort, sink) : nonOpCancel;
+    createPromiseCallback('sink.abort', abort, sink) :
+    nonOpCancel;
   setupWritableStreamDefaultController(
     stream,
     controller,
@@ -1356,4 +1358,5 @@ module.exports = {
   writableStreamDefaultControllerAdvanceQueueIfNeeded,
   setupWritableStreamDefaultControllerFromSink,
   setupWritableStreamDefaultController,
+  createWritableStream,
 };
