@@ -127,6 +127,10 @@ function normalizeString(path, allowAboveRoot, separator, isPathSeparator) {
   return res;
 }
 
+function formatExt(ext) {
+  return ext ? `${ext[0] === '.' ? '' : '.'}${ext}` : '';
+}
+
 /**
  * @param {string} sep
  * @param {{
@@ -142,7 +146,7 @@ function _format(sep, pathObject) {
   validateObject(pathObject, 'pathObject');
   const dir = pathObject.dir || pathObject.root;
   const base = pathObject.base ||
-    `${pathObject.name || ''}${pathObject.ext || ''}`;
+    `${pathObject.name || ''}${formatExt(pathObject.ext)}`;
   if (!dir) {
     return base;
   }
@@ -643,7 +647,7 @@ const win32 = {
       return `\\\\?\\${resolvedPath}`;
     }
 
-    return path;
+    return resolvedPath;
   },
 
   /**
