@@ -20,6 +20,8 @@
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 #include "env-inl.h"
+#include "node_external_reference.h"
+#include "permission/permission.h"
 #include "stream_base-inl.h"
 #include "stream_wrap.h"
 #include "util-inl.h"
@@ -66,6 +68,12 @@ class ProcessWrap : public HandleWrap {
     SetConstructorFunction(context, target, "Process", constructor);
   }
 
+  static void RegisterExternalReferences(ExternalReferenceRegistry* registry) {
+    registry->Register(New);
+    registry->Register(Spawn);
+    registry->Register(Kill);
+  }
+
   SET_NO_MEMORY_INFO()
   SET_MEMORY_INFO_NAME(ProcessWrap)
   SET_SELF_SIZE(ProcessWrap)
@@ -147,6 +155,8 @@ class ProcessWrap : public HandleWrap {
     Local<Context> context = env->context();
     ProcessWrap* wrap;
     ASSIGN_OR_RETURN_UNWRAP(&wrap, args.Holder());
+    THROW_IF_INSUFFICIENT_PERMISSIONS(
+        env, permission::PermissionScope::kChildProcess, "");
     int err = 0;
 
     Local<Object> js_options =
@@ -332,3 +342,5 @@ class ProcessWrap : public HandleWrap {
 }  // namespace node
 
 NODE_BINDING_CONTEXT_AWARE_INTERNAL(process_wrap, node::ProcessWrap::Initialize)
+NODE_BINDING_EXTERNAL_REFERENCE(process_wrap,
+                                node::ProcessWrap::RegisterExternalReferences)
