@@ -14,9 +14,10 @@ const server = createServer(common.mustCall((req, res) => {
   onGC(req, { ongc: common.mustCall(() => { server.close(); }) });
   req.resume();
   req.on('end', common.mustCall(() => {
-    setImmediate(() => {
+    setImmediate(async () => {
       client.end();
-      global.gc();
+      await global.gc({ type: 'major', execution: 'async' });
+      await global.gc({ type: 'major', execution: 'async' });
     });
   }));
   res.end('hello world');
@@ -32,7 +33,7 @@ server.listen(0, common.mustCall(() => {
     'Content-Length: 11',
     '',
     'hello world',
-    ''
+    '',
   ].join('\r\n');
 
   client.write(req);
