@@ -18,8 +18,8 @@ for (let i = 0; i < 1000000; i++) {
 async function main() {
   await common.gcUntil('RSS should go down', () => {
     const after = process.memoryUsage.rss();
-    if (process.config.variables.asan) {
-      console.log(`asan: before=${before} after=${after}`);
+    if (common.isASan) {
+      console.log(`ASan: before=${before} after=${after}`);
       return after < before * 10;
     } else if (process.config.variables.node_builtin_modules_path) {
       console.log(`node_builtin_modules_path: before=${before} after=${after}`);
